//
//  MYOttoman.h
//  Ottoman
//
//  Created by Jens Alfke on 9/21/09.
//  Copyright 2009 Jens Alfke. All rights reserved.
//

#import <Foundation/Foundation.h>

@class MYVersionDictionary, MYCurrentVersionDictionary;


/** A version-controlled persistent dictionary. 
    Each version is stored as a MYVersionDictionary, 
    unsaved changes as an MYCurrentVersionDictionary. */
@interface MYOttoman : NSObject 
{
    void *_ottoman;
    MYVersionDictionary *_lastVersion;
    MYCurrentVersionDictionary *_currentVersion;
}

/** Creates an "untitled" Ottoman with no file and no lastVersion.
    After adding values to the currentVersion, use saveAs to save it. */
- (id) init;

/** Opens an existing Ottoman file. */
- (id) initWithURL: (NSURL*)fileURL writeable: (BOOL)writeable error: (NSError**)outError;

/** Closes an Ottoman. */
- (void) close;

/** The current file, or nil if the receiver is still in the "untitled" state. */
@property (readonly) NSURL *URL;

/** The latest saved version of the dictionary. 
    Earlier versions can be accessed through its previousVersion property.
    This will be nil if the receiver is still in the "untitled" state. */
@property (readonly) MYVersionDictionary* lastVersion;

/** A mutable overlay representing the current state of the dictionary.
    Changes are made in memory until -save is called.
    This will be nil if the receiver was opened read-only (writeable=NO). */
@property (readonly) MYCurrentVersionDictionary* currentVersion;

/** Has the on-disk file been updated with newer revisions than what I know about? */
@property (readonly) BOOL needsSync;

/** Reads any newer versions from disk.
    Returns YES if new versions were read, NO if there were none or on error.
    Afterwards, -lastVersion will return the latest version in the file. 
    (The old lastVersion dictionary is still around, so you can save a pointer to it 
    before the call and use it later to see what changed.)
    Changes made to the -currentVersion dictionary are not lost, but are now relative
    to the new lastVersion. You may want to scan them and resolve conflicts. */
- (BOOL) sync: (NSError**)outError;

/** Saves the current version to the file, by appending.
    Returns NO if there is a version conflict; in that case you need to call -sync,
    possibly resolve conflicts, and then call -save again. */
- (BOOL) save: (NSError**)outError;

/** Saves the current version to the file, by writing to a new temporary file and
    then atomically replacing the original. 
    Older versions, and older copies of the data, are not preserved, so this
    will typically shrink the file quite a bit.
    Returns NO if there is a version conflict. */
- (BOOL) saveAndCompact: (NSError**)outError;

/** Saves the current version to a new file, leaving the new file open, 
    so subsequent saves will be written to it. */
- (BOOL) saveAs: (NSURL*)newFileURL
         overwriteAllowed: (BOOL)overwriteAllowed
         error: (NSError**)outError;

/** Scans the file for damage. Returns YES if the file is OK, NO if problems are found.
    If the 'repair' flag is set, will truncate the file to the last valid version. */
- (BOOL) scavengeAndRepair: (BOOL)repair error: (NSError**)outError;

@end
